part of masamune.dartcore;

/// Object extension methods
extension ObjectExtension on Object {
  /// Returns as a string
  String get asString => this as String;
  /// Returns as a int
  int get asInt => this as int;
  /// Returns as a double
  double get asDouble => this as double;
  /// Returns as a bool
  bool get asBool => this as bool;
}
/// String extension methods
extension StringExtension on String {
  /// Trim with specific characters
  /// 
  /// [chars]: The string to trim
  String trimString( String chars ){
    RegExp pattern = (chars != null) ? RegExp("^[$chars]+|[$chars]+\$") : RegExp(r'^\s+|\s+$');
    return this.replaceAll(pattern, '');
  }
  /// Trim only the left side with a specific string
  /// 
  /// [chars]: The string to trim
  String trimStringLeft( String chars ){
    RegExp pattern = (chars != null) ? RegExp("^[$chars]+") : RegExp(r'^\s+');
    return this.replaceAll(pattern, '');
  }
  /// Trim only the right side with a specific string
  /// 
  /// [chars]: The string to trim
  String trimStringRight( String chars ){
    RegExp pattern = (chars != null) ? RegExp("[$chars]+\$") : RegExp(r'\s+$');
    return this.replaceAll(pattern, '');
  }
  /// Converts a String to an int.
  /// 
  /// Normally it parses, but if it cannot parse it, it creates a random string.
  int toInt(){
    if( isEmpty(this) ) return 0;
    int i = int.tryParse(this);
    if( i != null ) return i;
    int val = 0;
    for( int rune in this.runes ){
        val += ( val + 1 ) * rune;
    }
    return val;
  }
  /// Converts a String to an int.
  /// 
  /// Normally it parses, but if it cannot parse it, it creates a random string.
  double toDouble(){
    if( isEmpty(this) ) return 0;
    double d = double.tryParse(this);
    if( d != null ) return d;
    double val = 0;
    for( int rune in this.runes ){
        val += ( val + 1 ) * rune;
    }
    return val;
  }
  /// Parses a String into a bool.
  bool toBool(){
    if( this.toLowerCase() == "true" ) return true;
    return false;
  }
  /// Convert to SHA1 hash
  String toSHA1() => sha1.convert(utf8.encode(this)).toString();
  /// Convert to SHA256 hash
  /// 
  /// [password]: Password
  String toSHA256( String password ){
    if( isEmpty(password) ) return this;
    Hmac hmacSha256 = Hmac(sha256, utf8.encode(password));
    return hmacSha256.convert(utf8.encode("foobar")).toString();
  }
}
/// Random extension methods
extension RandomExtension on Random {
  /// Get a random number from [min] to [max]
  /// 
  /// [min]: Minimum value
  /// [max]: Maximum value
  int rangeInt( int min, int max ) => ( ( this.nextDouble() * ( max - min ) ) + min ).toInt().limit(min, max);
  /// Get a random number from [min] to [max]
  /// 
  /// [min]: Minimum value
  /// [max]: Maximum value
  double rangeDouble( double min, double max ) => ( ( this.nextDouble() * ( max - min ) ) + min ).limit(min, max);
}
/// Int extension methods
extension IntExtention on int {
  /// Restrict value from [min] to [max]
  /// 
  /// [min]: Minimum value
  /// [max]: Maximum value
  int limit( int min, int max ){
    if( this < min ) return min;
    if( this > max ) return max;
    return this;
  }
  /// Restrict value from [min]
  /// 
  /// [min]: Minimum value
  int limitLow( int min ){
    if( this < min ) return min;
    return this;
  }
  /// Restrict value from [max]
  /// 
  /// [max]: Maximum value
  int limitHigh( int max ){
    if( this > max ) return max;
    return this;
  }
}
/// Double extension methods
extension DoubleExtention on double {
  /// Restrict value from [min] to [max]
  /// 
  /// [min]: Minimum value
  /// [max]: Maximum value
  double limit( double min, double max ){
    if( this < min ) return min;
    if( this > max ) return max;
    return this;
  }
  /// Restrict value from [min]
  /// 
  /// [min]: Minimum value
  double limitLow( double min ){
    if( this < min ) return min;
    return this;
  }
  /// Restrict value from [max]
  /// 
  /// [max]: Maximum value
  double limitHigh( double max ){
    if( this > max ) return max;
    return this;
  }
}
/// Bool extension methods
extension BoolExtension on bool {
  /// Compares two values
  /// 
  /// [bool]: Other values
  int compareTo( bool other ){
    if( this == other ) return 0;
    if( this == true ) return 1;
    if( other == true ) return -1;
    return 0;
  }
}
/// DateTime extension methods
extension DateTimeExtension on DateTime {
  /// Gets the time per frame in UnixTime in milliseconds.
  int get frameMillisecondsSinceEpoch => ( this.millisecondsSinceEpoch / 10 ).round() * 10;
}
/// Future extension methods
extension FutureExtension on Future {
  /// Check if this Future has finished
  /// True if finished
  bool get isCompleted {
    bool isCompleted = false;
    this.whenComplete(() => isCompleted = true );
    return isCompleted;
  }
}