part of masamune.data;

/// Collection class for storing data
/// Can store DataDocument.
/// Since it has the same structure as FireStore, it can be used without discomfort.
/// [Iterable<DataDocument>] is available in the for statement because it is MixIn.
/// ```
/// DataCollection collection = DataCollection();
/// collection.add( DataDocument.fromMap( "data/document", {"A": "Data is A", "B": "Data is B"} ) );
/// for( DataDocument document in colletion ){ ... }
/// ```
class DataCollection extends Collection<DataDocument> implements IDataCollection<DataDocument> {
  static Iterable<DataDocument> _convertDataFromMap( String path, Map<String, dynamic> data ) {
    List<DataDocument> list = ListPool.get();
    data?.forEach(
      ( key, value ){        
        if( isEmpty(key) || value == null ) return;
        if( !( value is Map ) ) return;
        list.add(DataDocument.fromMap(Paths.child(path, key), (value as Map )));
      });
    return list;
  }
  static Iterable<DataDocument> _convertDataFromList( String path, Iterable<dynamic> data ) {
    List<DataDocument> list = ListPool.get();   
    data?.forEachIndex(
      ( value, i ){
      if( value == null || !( value is Map ) ) return;
      list.add(DataDocument.fromMap(Paths.child(path, i.toString()), (value as Map )));
    });
    return list;
  }
  
  /// Process to create a new instance
  /// Do not use from outside the class
  @override
  @protected
  T createInstance<T extends IClonable>() => DataCollection( ) as T;
  /// Collection class for storing data
  /// Can store DataDocument.
  /// Since it has the same structure as FireStore, it can be used without discomfort.
  /// [Iterable<DataDocument>] is available in the for statement because it is MixIn.
  /// ```
  /// DataCollection collection = DataCollection();
  /// collection.add( DataDocument.fromMap( "data/document", {"A": "Data is A", "B": "Data is B"} ) );
  /// for( DataDocument document in colletion ){ ... }
  /// ```
  /// [path]: Collection path
  factory DataCollection([String path]){
    path = path?.trimString( Const.slash )?.replaceAll( "//", "/" )?.applyTags();
    if( isNotEmpty(path) ) {
      DataCollection collection = PathMap.get<DataCollection>( path );
      if( collection != null ) return collection;
    }
    return DataCollection._( path:path );
  }
  /// Collection class for storing data
  /// Can store DataDocument.
  /// Since it has the same structure as FireStore, it can be used without discomfort.
  /// [Iterable<DataDocument>] is available in the for statement because it is MixIn.
  /// ```
  /// DataCollection collection = DataCollection();
  /// collection.add( DataDocument.fromMap( "data/document", {"A": "Data is A", "B": "Data is B"} ) );
  /// for( DataDocument document in colletion ){ ... }
  /// ```
  /// [path]: Collection path
  /// [jsonData]: Json string representing Map
  factory DataCollection.fromJson( String path, String jsonData ){
    dynamic data = Json.decode(jsonData);
    switch( data?.runtimeType ){
      case List:
        return DataCollection.fromList( path, data as List );
      case Map:
        return DataCollection.fromMap( path, data as Map );
      default:
        return DataCollection( path );
    }
  }
  /// Collection class for storing data
  /// Can store DataDocument.
  /// Since it has the same structure as FireStore, it can be used without discomfort.
  /// [Iterable<DataDocument>] is available in the for statement because it is MixIn.
  /// ```
  /// DataCollection collection = DataCollection();
  /// collection.add( DataDocument.fromMap( "data/document", {"A": "Data is A", "B": "Data is B"} ) );
  /// for( DataDocument document in colletion ){ ... }
  /// ```
  /// [path]: Collection path
  /// [mapData]: Map data to be stored
  factory DataCollection.fromMap( String path, Map<String, dynamic> mapData ){
    path = path?.trimString( Const.slash )?.replaceAll( "//", "/" )?.applyTags();
    if( isNotEmpty(path) ) {
      DataCollection collection = PathMap.get<DataCollection>( path );
      if( collection != null ) {
          if( mapData != null ) collection.set( _convertDataFromMap( path, mapData ) );
          return collection;
      }
    }
    return DataCollection._( path:path, children:_convertDataFromMap( path, mapData ) );
  }
  /// Collection class for storing data
  /// Can store DataDocument.
  /// Since it has the same structure as FireStore, it can be used without discomfort.
  /// [Iterable<DataDocument>] is available in the for statement because it is MixIn.
  /// ```
  /// DataCollection collection = DataCollection();
  /// collection.add( DataDocument.fromMap( "data/document", {"A": "Data is A", "B": "Data is B"} ) );
  /// for( DataDocument document in colletion ){ ... }
  /// ```
  /// [path]: Collection path
  /// [listData]: List data to be stored
  factory DataCollection.fromList( String path, Iterable<dynamic> listData ){
    path = path?.trimString( Const.slash )?.replaceAll( "//", "/" )?.applyTags();
    if( isNotEmpty(path) ) {
      DataCollection collection = PathMap.get<DataCollection>( path );
      if( collection != null ) {
          if( listData != null ) collection.set( _convertDataFromList( path, listData ) );
          return collection;
      }
    }
    return DataCollection._( path:path, children:_convertDataFromList( path, listData ) );
  }
  DataCollection._({String path, Iterable<DataDocument> children}) : super( path, children: children );

  /// Order for sorting
  OrderBy orderBy;
  /// Compare key for sorting
  String orderByKey;
  /// Order for sorting
  /// Apply when the first sort is equivalent
  OrderBy thenBy;
  /// Compare key for sorting
  /// Apply when the first sort is equivalent
  String thenByKey;

  
  /// Sort according to the sort setting
  /// [orderBy] and [orderByKey], [thenBy] and [thenByKey] as the sort contents.
  void sort( ) => this.data.sort( this._sortAction );
  int _sortAction( MapEntry<String, DataDocument> a, MapEntry<String, DataDocument> b ) {
    int r = 0;
    if( a.value != null && b.value != null && a.value.containsKey( this.orderByKey ) && b.value.containsKey( this.orderByKey ) ) {
      if( this.orderBy == OrderBy.Asc ) {
        if( a.value[this.orderByKey] is double ) {
          r = a.value.getDouble( this.orderByKey ).compareTo( b.value.getDouble( this.orderByKey ) );
        } else if( a.value[this.orderByKey] is int ) {
          r = a.value.getInt( this.orderByKey ).compareTo( b.value.getInt( this.orderByKey ) );
        } else if( a.value[this.orderByKey] is bool ) {
          r = a.value.getBool( this.orderByKey ).compareTo( b.value.getBool( this.orderByKey ) );
        } else {
          r = a.value.getString( this.orderByKey ).compareTo( b.value.getString( this.orderByKey ) );
        }
      } else {
        if( a.value[this.orderByKey] is double ) {
          r = b.value.getDouble( this.orderByKey ).compareTo( a.value.getDouble( this.orderByKey ) );
        } else if( a.value[this.orderByKey] is int ) {
          r = b.value.getInt( this.orderByKey ).compareTo( a.value.getInt( this.orderByKey ) );
        } else if( a.value[this.orderByKey] is bool ) {
          r = b.value.getBool( this.orderByKey ).compareTo( a.value.getBool( this.orderByKey ) );
        } else {
          r = b.value.getString( this.orderByKey ).compareTo( a.value.getString( this.orderByKey ) );
        }
      }
    } else if( a.value == null || !a.value.containsKey( this.orderByKey ) ) {
      if( this.orderBy == OrderBy.Asc ) {
        r = -1;
      } else {
        r = 1;
      }
    } else if( b.value == null || !b.value.containsKey( this.orderByKey ) ) {
      if( this.orderBy == OrderBy.Asc ) {
        r = 1;
      } else {
        r = -1;
      }
    }
    if( r == 0 && this.thenBy != OrderBy.None && !isEmpty( this.thenByKey ) ) {
      if( a.value != null && b.value != null && a.value.containsKey( this.thenByKey ) && b.value.containsKey( this.thenByKey ) ) {
        if( this.thenBy == OrderBy.Asc ) {
          if( a.value[this.thenByKey] is double ) {
            r = a.value.getDouble( this.thenByKey ).compareTo( b.value.getDouble( this.thenByKey ) );
          } else if( a.value[this.thenByKey] is int ) {
            r = a.value.getInt( this.thenByKey ).compareTo( b.value.getInt( this.thenByKey ) );
          } else if( a.value[this.thenByKey] is bool ) {
            r = a.value.getBool( this.thenByKey ).compareTo( b.value.getBool( this.thenByKey ) );
          } else {
            r = a.value.getString( this.thenByKey ).compareTo( b.value.getString( this.thenByKey ) );
          }
        } else {
          if( a.value[this.thenByKey] is double ) {
            r = b.value.getDouble( this.thenByKey ).compareTo( a.value.getDouble( this.thenByKey ) );
          } else if( a.value[this.thenByKey] is int ) {
            r = b.value.getInt( this.thenByKey ).compareTo( a.value.getInt( this.thenByKey ) );
          } else if( a.value[this.thenByKey] is bool ) {
            r = b.value.getBool( this.thenByKey ).compareTo( a.value.getBool( this.thenByKey ) );
          } else {
            r = b.value.getString( this.thenByKey ).compareTo( a.value.getString( this.thenByKey ) );
          }
        }
      } else if( a.value == null || !a.value.containsKey( this.thenByKey ) ) {
        if( this.thenBy == OrderBy.Asc ) {
          r = -1;
        } else {
          r = 1;
        }
      } else if( b.value == null || !b.value.containsKey( this.thenByKey ) ) {
        if( this.thenBy == OrderBy.Asc ) {
          r = 1;
        } else {
          r = -1;
        }
      }
    }
    return r;
  }

  /// Get the data
  /// 
  /// [key]: Key for retrieving data
  DataDocument operator [](Object key) {
    if( key is int ){
      if( key < 0 || this.data.length <= key ) return null;
      return this.data[key];
    } else if( key is String ){
      if( isEmpty( key ) || !this.data.containsKey(key) ) return null;
      return this.data[key];
    }
    return null;
  }

}