part of masamune.path;

/// Attach a data filter to the IPath object
/// It is possible to process data by passing a filter when inputting and outputting data
/// ```
/// PathFilter.insertGetFilter( "data/user",  ( data ) => data + " is user's data." );
/// PathFilter.insertSetFilter( "data/user",  ( data ) => data.data = "Processed" );
/// ```
class PathFilter {
  static List _onGetFilter = ListPool.get();
  static List _onSetFilter = ListPool.get();
  static Map _processingCache = MapPool.get();

  /// Get the temporary data in the filter.
  /// 
  /// [path]: Path to get temporary data
  static dynamic getTemporary( String path ) {
    if( isEmpty( path ) || !_processingCache.containsKey( path ) ) return null;
    return _processingCache[path];
  }
  /// Get the temporary data in the filter.
  /// 
  /// [path]: Path to get temporary data
  static bool setTemporary( String path, dynamic value ) {
    if( isEmpty( path ) || !_processingCache.containsKey( path ) ) return false;
    return ( _processingCache[path] = value ) != null;
  }
  /// Apply a get filter for a specific path.
  /// By specifying a region, you can apply a filter that is limited to the region.
  /// 
  /// [path]: The path to apply the filter to
  /// [value]: Data to filter
  /// [region]: Region to apply the filter to
  static TValue applyGetFilter<TValue>( String path, TValue value, {String region} ) {
    if( isEmpty( path ) ) return value;
    _processingCache[path] = null;
    for( int i = _onGetFilter.length - 1; i >= 0; i-- ) {
      _FilterPath tmp = _onGetFilter[i];
      if( tmp.path != path )continue;
      if( isNotEmpty(region) && region != tmp.region ) continue;
      if( tmp.action == null || ( tmp.observer != null && tmp.observer.isDisposed ) ) {
          _onGetFilter.removeAt( i );
          continue;
      }
      TValue source = value;
      _processingCache[path] = null; 
      value = tmp.action( value );
      if( source == value && _processingCache.containsKey( path ) && _processingCache[path] != null ) value = _processingCache[path];
    }
    _processingCache.remove( path );
    return value;
  }
  /// Apply a set filter for a specific path.
  /// By specifying a region, you can apply a filter that is limited to the region.
  /// 
  /// [path]: The path to apply the filter to
  /// [value]: Data to filter
  /// [region]: Region to apply the filter to
  static TValue applySetFilter<TValue>( String path, TValue value, { String region} ) {
    if( isEmpty(path) ) return value;
    _processingCache[path] = null;
    for( int i = _onSetFilter.length- 1; i >= 0; i-- ) {
      _FilterPath tmp = _onSetFilter[i];
      if( tmp.path != path ) continue;
      if( isNotEmpty(region) && region != tmp.region ) continue;
      if( tmp.action == null || ( tmp.observer != null && tmp.observer.isDisposed ) ) {
          _onSetFilter.removeAt( i );
          continue;
      }
      TValue source = value;
      _processingCache[path] = null;
      value = tmp.action( value );
      if( source == value && _processingCache.containsKey( path ) && _processingCache[path] != null ) value = _processingCache[path];
    }
    _processingCache.remove( path );
    return value;
  }
  /// Inserts GetFilter into the specified path
  /// You can specify the execution order
  /// By specifying [observer], it is possible to monitor only while the object exists
  /// 
  /// [path]: The path to filter
  /// [filter]: Content to filter
  /// [region]: Filter region
  /// [order]: Filter execution order, smaller is faster
  /// [observer]: Monitoring object. If this is deleted, the filter will be deleted automatically
  static void insertGetFilter( String path, dynamic filter( dynamic value ), {String region, int order = 0, Observer observer}) {
    if( isEmpty(path) || filter == null ) return;
    int hash = filter.hashCode;
    if( _containsOnGetFilter( path, hash, region ) ) return;
    int position = 0;
    for( int i = 0; i < _onGetFilter.length; i++ ) {
      if( _onGetFilter[i].order < order ) break;
      position = i;
    }
    Log.msg( "Adding to PathFilter: %s at %s", [path, position]);
    _onGetFilter.insert( position, _FilterPath( path, region, hash, filter, order, observer ) );
  }
  /// Removes a filter from the specified path
  /// 
  /// [path]: The path to delete
  /// [filter]: The filter to delete. Delete only matches
  /// [region]: Filter region
  static void removeGetFilter( String path, dynamic filter( dynamic value ), {String region} ) {
    if( isEmpty(path) || filter == null ) return;
    int hash = filter.hashCode;
    for( int i = _onGetFilter.length - 1; i >= 0; i-- ) {
      _FilterPath tmp = _onGetFilter[i];
      if( path != tmp.path || hash != tmp.hash || region != tmp.region ) continue;
      _onGetFilter.removeAt( i );
    }
  }
  /// Inserts SetFilter into the specified path
  /// You can specify the execution order
  /// By specifying [observer], it is possible to monitor only while the object exists
  /// 
  /// [path]: The path to filter
  /// [filter]: Content to filter
  /// [region]: Filter region
  /// [order]: Filter execution order, smaller is faster
  /// [observer]: Monitoring object. If this is deleted, the filter will be deleted automatically
  static void insertSetFilter( String path, dynamic filter( dynamic value ), {String region, int order = 0, Observer observer} ) {
    if( isEmpty(path) || filter == null ) return;
    int hash = filter.hashCode;
    if( _containsOnSetFilter( path, hash, region ) ) return;
    int position = 0;
    for( int i = 0; i < _onSetFilter.length; i++ ) {
      if( _onSetFilter[i].order < order ) break;
      position = i;
    }
    Log.msg( "Adding to PathFilter: %s at %s", [path, position] );
    _onSetFilter.insert( position, _FilterPath( path, region, hash, filter, order, observer ) );
  }
  /// Removes a filter from the specified path
  /// 
  /// [path]: The path to delete
  /// [filter]: The filter to delete. Delete only matches
  /// [region]: Filter region
  static void removeSetFilter( String path, dynamic filter( dynamic value ), {String region} ) {
    if( isEmpty(path) || filter == null ) return;
    int hash = filter.hashCode;
    for( int i = _onSetFilter.length - 1; i >= 0; i-- ) {
      _FilterPath tmp = _onSetFilter[i];
      if( path != tmp.path || hash != tmp.hash || region != tmp.region ) continue;
      _onSetFilter.removeAt( i );
    }
  }
  static bool _containsOnGetFilter( String path, int hash, String region ) {
    if( isEmpty(path) ) return false;
    for( _FilterPath tmp in _onGetFilter ){
      if( tmp.path != path || hash != tmp.hash || region != tmp.region ) continue;
      return true;
    }
    return false;
  }
  static bool _containsOnSetFilter( String path, int hash, String region ) {
    if( isEmpty(path) ) return false;
    for( _FilterPath tmp in _onSetFilter ) {
      if( tmp.path != path || hash != tmp.hash || region != tmp.region ) continue;
      return true;
    }
    return false;
  }
  /// Clear all filters
  static void clear(){
    _onGetFilter.clear();
    _onSetFilter.clear();
    _processingCache.clear();
  }
}  
class _FilterPath {
      String path;
      String region;
      int hash;
      dynamic Function(dynamic value) action;
      int order;
      Observer observer;
      _FilterPath( String path, String reegion, int hash, dynamic action(dynamic value), int order, Observer observer ) {
         this.path = path;
         this.region = region;
         this.hash = hash;
         this.action = action;
         this.order = order;
         this.observer = observer;
      }
   }