part of masamune.path;

/// Monitor filter changes
/// Specify the monitoring target path in [listen()] and execute the process when there is any change
/// ```
/// PathListener.listen( "data/user", ( path ) => Log.msg( "Updated" ) );
/// ```
class PathListener {
  static Timer _timer;
  static List _listener = ListPool.get();
  static List _updateStack = ListPool.get();
  static void _startUpdate( ){
    if( _timer != null ) return;
    _timer = Timer.periodic(
      Config.frameTime, 
      (timer) {
        List<IPath> cache = ListPool.get( );
        while( _updateStack.length > 0 ) {
          IPath path = _updateStack.removeLast( );
          if( path == null || path.isTemporary || isEmpty( path.path ) || cache.contains( path ) ) continue;
          cache.add( path );
          _invokeListener( path.path );
        }
        cache.release( );
      }
    );
  }

  /// Start monitoring for path changes.
  /// If there is a change in [path], [onUpdate] is called.
  /// If [observer] is set, monitoring will automatically end when [observer] is deleted
  /// 
  /// [path]: Path to monitor
  /// [onUpdate]: Callback executed when [path] changes
  /// [observer]: Monitoring object. If this is deleted, monitoring will automatically end
  static void listen<T extends IPath>( String path, void onUpdate( T path ), {Observer observer} ) {    
    if( isEmpty(path) || onUpdate == null ) return;
    _startUpdate();
    Type type = T;
    int hash = onUpdate.hashCode;
    if( _containsListener( path: path, type:type, hash:hash ) ) return;
    Log.msg( "Adding to PathListener: %s and %s", [path, T] );
    _listener.add( new _ListenerPath( path, type, hash, ( r ) => onUpdate( r as T ), observer ) );
  }
  /// Stop monitoring a specific path
  /// 
  /// [path]: Path to monitor
  /// [onUpdate]: Processing to stop monitoring. Exit only matches
  static void unlisten<T extends IPath>( String path, void onUpdate( T path ) ) {
    if( isEmpty(path) || onUpdate == null ) return;
    Type type = T;
    int hash = onUpdate.hashCode;
    for( int i = _listener.length - 1; i >= 0; i-- ) {
      _ListenerPath tmp = _listener[i];
      if( tmp.path != path || type != tmp.type || tmp.hash != hash ) continue;
      Log.msg( "Removing from PathListener: %s and %s", [path, T] );
      _listener.removeAt( i );
    }
  }
  static void _updateNotify( IPath path ) {
    Log.msg( "Enqueue to invoking action: %s (%s)", [path.path, path.toString()] );
    _updateStack.add( path );
  }
  static void _invokeListener( String path ) {
    if( isEmpty( path ) ) return;
    for( int i = _listener.length - 1; i >= 0; i-- ) {
      _ListenerPath tmp = _listener[i];
      if( tmp.path != path ) continue;
      if( tmp.action == null || ( tmp.observer != null && tmp.observer.isDisposed ) ) {
          _listener.removeAt( i );
          continue;
      }
      IPath val = PathMap.get( path );
      if( val == null ) continue;
      tmp.action( val );
    }
  }
  static bool _containsListener( {String path, Type type, int hash} ) {
    for( _ListenerPath tmp in _listener ) {
      if( ( isNotEmpty(path) && tmp.path != path ) || tmp.type != type || hash != tmp.hash ) continue;
      return true;
    }
    return false;
  }
  /// Clear all listeners
  static void clear(){
    _listener.clear();
    _updateStack.clear();
  }
}
class _ListenerPath {
  String path;
  Type type;
  int hash;
  void Function( IPath path ) action;
  Observer observer;
  _ListenerPath( String path, Type type, int hash, void Function( IPath path ) action, Observer observer ) {
    this.path = path;
    this.type = type;
    this.hash = hash;
    this.action = action;
    this.observer = observer;
  }
}
