part of masamune.path;

/// Manages pal system paths
/// Map all non-temporary path data to avoid duplication
/// Parental and child relationships can be maintained, grouped, and collectively deleted by type
/// ```
/// PathMap.get( "data/asset" );
/// ```
class PathMap {
  static List<IPath> _data = ListPool.get();
  static Map<String,String> _link = MapPool.get();

  /// Gets the object at the specified path
  /// If no path is specified, search for a matching type from all paths and return the first object that matches
  /// 
  /// [path]: The path to get
  static T get<T extends Object>([String path]) {
    for( IPath tmp in _data ) {
      if( tmp == null || ( isNotEmpty( path ) && tmp.path != path ) ) continue;
      if( tmp is IUnit &&  tmp .data is T ) return tmp.data;
      if( !( tmp is T ) ) { Log.error( "The object type is different: %s and %s", [tmp.runtimeType, T] ); return null; }
      return tmp as T;
    }
    if( isEmpty(path) ) return null;
    for( IPath tmp in _data ) {
      if( tmp == null || tmp.path == path || !( tmp is IUnit ) || !path.startsWith(tmp.path) ) continue;
      if( !( (tmp as IUnit)[path.replaceAll(tmp.path, Const.empty)] is T ) ) { Log.error( "The object type is different: %s and %s", [tmp.runtimeType, T] ); return null; }
      return tmp as T;
    }
    if( _link.containsKey(path) ) return get<T>( _link[path] );
    return null;
  }
  /// Gets all path objects of the matching type from the pathmap
  static List<T> getAll<T extends IPath>( ) {
    List<T> list = ListPool.get( );
    for( IPath tmp in _data ) {
      if( tmp == null ) continue;
      if( !( tmp is T ) ) continue;
      list.add( tmp as T );
    }
    return list;
  }
  /// Check if the path exists in the pathmap
  /// True if the path exists in the pathmap
  /// 
  /// [path]: The path to check
  static bool containsKey( String path ) {
    if( isEmpty(path) ) return false;
    if( _link.containsKey(path) ) return true;
    for( IPath tmp in _data ) {
      if( tmp == null || tmp.path != path ) continue;
      return true;
    }
    return false;
  }
  static bool _containsKey( IPath path, Type type ) {
    if( path == null || isEmpty(path.path) ) return false;
    for( IPath tmp in _data ) {
      if( tmp == null || tmp.path != path.path || tmp.runtimeType != type || tmp.protocol != path.protocol ) continue;
      return true;
    }
    return false;
  }
  /// Check if the path object exists in the pathmap
  /// True if the path object exists in the pathmap
  /// 
  /// [value]: The path object to check
  static bool containsValue( IPath value ) {
    if( value == null ) return false;
    for( IPath tmp in _data ) {
      if( value != tmp ) continue;
      return true;
    }
    return false;
  }
  /// Add a new path object to the pathmap
  /// Temporary data or duplicate additions are not allowed
  /// Also, if there is a parent-child relationship with the existing path, it will be adjusted automatically
  /// 
  /// [data]: The path object to add
  static void add( IPath data ) {
    if( data == null ) {
      Log.error( "Adding data is invalid." );
      return;
    }
    if( data.isTemporary ) {
      Log.error( "Can't add the temporary data. please check the IsTemporary value" );
      return;
    }
    Log.msg( "Adding to PathMap: %s", [data.path] );
    if( PathMap._containsKey( data, data.runtimeType ) ) {
      Log.error( "The data path is already exist: %s and %s", [data.runtimeType, data.path] );
    }
    _data.add( data );
    if( data is IChild ) {
      for( IPath path in _data ) {
          if( !( path is IParent ) ) continue;
          if( Paths.parent( data.path ) != path.path ) continue;
          ( path as IParent )._addChildInernal( data );
      }
    }
  }
  /// Removes an object from the pathmap given its path
  /// 
  /// [path]: The path to delete
  static void removePath( String path ) {
    if( isEmpty(path) ) return;
    for( int i = _data.length- 1; i >= 0; i-- ) {
      IPath data = _data[i];
      if( data == null || data.path != path || data.group < 0 || !data.isDisposable ) continue;
      _removeInternal( data );
      data.dispose( );
      _data.removeAt( i );
    }
    _link.remove(path);
  }
  /// Removes all corresponding path objects from the specified path list
  /// 
  /// [paths]: The path list to delete
  static void removeAllPath( Iterable<String> paths ) {
      if( paths == null ) return;
      for( String key in paths ) removePath( key );
  }
  /// Removes an object from the pathmap given its group number
  /// 
  /// [group]: The group number to delete
  static void removeGroup( int group ) {
    for( int i = _data.length - 1; i >= 0; i-- ) {
      IPath data = _data[i];
      if( data == null || data.group != group || !data.isDisposable ) continue;
      _removeInternal( data );
      data.dispose( );
      _data.removeAt( i );
    }
  }
  /// Removes all corresponding path objects from the specified group list
  /// 
  /// [groups]: The group list to delete
  static void removeAllGroup( Iterable<int> groups ) {
    if( groups == null ) return;
    for( int key in groups ) removeGroup( key );
  }
  /// Deletes the specified path object
  /// 
  /// [path]: The path object to delete
  static void remove( IPath path ) {
    if( path == null || isEmpty(path.path) ) return;
    for( int i = _data.length - 1; i >= 0; i-- ) {
      IPath data = _data[i];
      if( data == null || path != data || data.group < 0 || !data.isDisposable ) continue;
      _removeInternal( data );
      data.dispose( );
      _data.removeAt( i );
    }
  }
  static void _removeInternal( IPath path ) {
    if( !( path is IChild ) ) return;
    for( IPath data in _data ) {
      if( !( data is IParent ) ) continue;
      if( Paths.parent( path.path ) != data.path ) continue;
      ( data as IParent )._removeChildInternal( path );
    }
  }
  /// Removes an object from the pathmap given its type
  /// 
  /// [group]: The type to delete
  static void removeType( Type type ) {
    if( type == null ) return;
    for( int i = _data.length - 1; i >= 0; i-- ) {
      IPath data = _data[i];
      if( data == null || data.runtimeType != type || data.group < 0 || !data.isDisposable ) continue;
      _removeInternal( data );
      data.dispose( );
      _data.removeAt( i );
    }
  }
  /// Removes all corresponding path objects from the specified type list
  /// 
  /// [groups]: The type list to delete
  static void removeAllType( Iterable<Type> types ) {
    if( types == null ) return;
    for( Type key in types ) removeType( key );
  }
  /// Deletes path objects that match the specified conditions
  /// 
  /// [predicate]: The condition to delete. If [predicate] is True, delete the object
  static void removeOn( bool predicate( Type type, String path, IPath item ) ) {
    if( predicate == null ) return;
    try {
      for( int i = _data.length - 1; i >= 0; i-- ) {
          IPath data = _data[i];
          if( data == null || data.group < 0 || !data.isDisposable || !predicate( data.runtimeType, data.path, data ) ) continue;
          _removeInternal( data );
          data.dispose( );
          _data.removeAt( i );
      }
    } catch(e) { Log.error( e ); }
  }
  /// Callback event when application pauses
  /// 
  /// [pause]: True when application is paused, False when returning from pause
  static void onApplicationPause( bool pause ) {
    for( int i = _data.length - 1; i >= 0; i-- ) {
      IPath data = _data[i];
      if( data == null ) continue;
      data.onApplicationPause( pause );
    }
  }
  /// Callback event when application quit
  static void onApplicationQuit( ) {
    for( int i = _data.length - 1; i >= 0; i-- ) {
      IPath data = _data[i];
      if( data == null ) continue;
      data.onApplicationQuit( );
    }
  }
  /// Clear all pathmaps
  static void clear(){
    _data.clear();
  }

  /// Link with another path
  /// You can access the same object even if you access the linked path
  /// If the path of the actual object exists, it takes precedence
  /// If the link destination is a link, it will be overwritten.
  /// 
  /// [fromPath]: Referrer path
  /// [toPath]: Referenced path
  static void link( String fromPath, String toPath ){
    if( isEmpty(fromPath) || isEmpty(toPath) ){
      Log.error( "The path is invalid." );
      return;
    }
    _link[toPath] = fromPath;
  }

}