part of masamune.task;

/// Collection class with task management function
/// The task performed is usually the [Future<T>] is returned.
/// Execution is completed when [isDone] becomes True.
/// If an error occurs, [isError] is True and the associated error flag is True.
/// Everything is built into the Path system and treated as a Collection.
/// 
/// Please inherit and use
abstract class TaskCollection<TChild extends IChild> extends Collection<TChild> implements ITask {
  Completer _completer;
  /// Create a Completer that matches the class.
  /// Do not use from external class
  @protected
  Completer createCompleter() => Completer<Task>();
  /// Returns the [Future] class of the task.
  /// Normally, use it inside a class. Do not use outside the class.
  @protected
  Future get future => this._completer?.future;

  
  /// Collection class with task management function
  /// The task performed is usually the [Future<T>] is returned.
  /// Execution is completed when [isDone] becomes True.
  /// If an error occurs, [isError] is True and the associated error flag is True.
  /// Everything is built into the Path system and treated as a Collection.
  /// 
  /// [path]: Document path
  /// [children]: Initial document data, using TChild array
  /// [group]: Unit group (0 or more)
  /// [order]: Unit order
  @protected
  TaskCollection({String path, Iterable<TChild> children, int group = 0, int order = 10}) : super( path, children:children, group:group, order:order ){
    this._completer = this.createCompleter();
  }

  /// Get the data
  /// 
  /// [key]: Key for retrieving data
  TChild operator []( Object key ) {
    if( key is int ){
      if( key < 0 || this.data.length <= key ) return null;
      return this.data[key];
    } else if( key is String ){
      if( isEmpty( key ) || !this.data.containsKey(key) ) return null;
      return this.data[key];
    }
    return null;
  }
   
  /// True when the object has finished executing
  bool get isDone => this._isDone;
  bool _isDone = false;
  /// True if the object is in error
  bool get isError => this._isError;
  bool _isError = false;
  /// True if object or application is restricted
  bool get isRestricted => this._isRestricted;
  @protected
  set isRestricted( bool isRestricted ) => this._isRestricted = isRestricted;
  bool _isRestricted = false;
  /// True if not connected to network
  bool get isNotReachableNetwork => this._isNotReachableNetwork;
  @protected
  set isNotReachableNetwork( bool isNotReachableNetwork ) => this._isNotReachableNetwork = isNotReachableNetwork;
  bool _isNotReachableNetwork = false;
  /// True when system is not supported
  bool get isNotSupported => this._isNotSupported;
  @protected
  set isNotSupported( bool isNotSupported ) => this._isNotSupported = isNotSupported;
  bool _isNotSupported = false;
  /// True if object or data is duplicated
  bool get isDuplicate => this._isDuplicate;
  @protected
  set isDuplicate( bool isDuplicate ) => this._isDuplicate = isDuplicate;
  bool _isDuplicate = false;
  /// True if object or data not found
  bool get isNotFound => this._isNotFound;
  @protected
  set isNotFound( bool isNotFound ) => this._isNotFound = isNotFound;
  bool _isNotFound = false;
  /// Get error message
  String get message => this._message;
  String _message;
  /// Gets the progress（0.0 - 1.0）.
  double get progress => this._progress;
  @protected
  set progress( double progress ) => this._progress = progress;
  double _progress = 0;

  /// Initialize execution
  /// You can use it publicly, but do not use it from outside the class.
  @protected
  T init<T extends IDoable>( ) {
    if( this._completer.isCompleted ) this._completer = this.createCompleter();
    this._isDone = this._isError = this._isAbort = false;
    this._isNotSupported = this._isDuplicate = this._isRestricted = this._isNotFound = this._isNotReachableNetwork = false;
    this.isCloned = false;
    this.progress = 0;
    this._message = null;
    return this as T;
  }
  /// Exit as successful.
  /// You can use it publicly, but do not use it from outside the class.
  @protected
  T done<T extends IDoable>(){
    if( this.isDone ) return this as T;
    this._isDone = true;
    this._isError = this._isAbort = false;
    this.progress = 1; this._message = null;
    this._completer.complete( this );
    this.notifyUpdate( );
    return this as T;
  }
  /// Terminates processing with an error.
  /// You can use it publicly, but do not use it from outside the class.
  /// 
  /// [message]: Error message
  /// [errorLevel]: Set the error level.
  @protected
  T error<T extends IErrorable>( String message, {ErrorLevel errorLevel = ErrorLevel.Error} ){
    this._isDone = this._isError = true;
    this._isAbort = false;
    this._message = message;
    this.progress = 1;
    this._errorLevel = errorLevel;
    if( errorLevel == ErrorLevel.Warning ) Log.warning( message );
    else if( errorLevel == ErrorLevel.Error ) Log.error( message );
    this._completer.complete( this );
    return this as T;
  }
  /// Get error level
  ErrorLevel get errorLevel => this._errorLevel;
  ErrorLevel _errorLevel = ErrorLevel.None;


  /// True if the object was interrupted
  bool get isAbort => this._isAbort;
  bool _isAbort = false;
  /// Abort the object
  /// Displays the abort [message].
  /// 
  /// [message]: Interruption message
  T abort<T extends IAbortable>([String message = Const.empty]) {
    this._isAbort = this._isDone = true;
    this._isError = false;
    this._message = message;
    this._completer.complete( this );
    return this as T;
  }


  /// True if the object is cached
  bool get isCached => false;
  /// True if object is locked
  bool get isLock => false;


  /// Debug output The debug output differs depending on the object
  /// 
  /// [value]: The object to enter
  T debug<T extends IDebuggable>([Object value]) {
    Log.msg( "Unit[%s][%s]\n%s", [this.runtimeType, this.path, this.toString( )] );
    return this as T;
  }
}